/*
 * Copyright 2008 Perforce Software.  All rights reserved.
 *
 * This file is part of Perforce - the FAST SCM System.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "MyDTS.h"
extern "C" {
#include "dtg-utils.h"
}
#include "dtg-str.h"

void print_DTGStrList( struct DTGStrList *list )
{
	struct DTGStrList *cur = list;
	while( cur )
	{
	    struct DTGStrList *next = cur->next;
	    if( cur->value )
	        printf( "cur->value:  \"%s\"\n", cur->value );
	    cur = next;
	}
}

int in_DTGStrList( struct DTGStrList *list, const char* str )
{
	struct DTGStrList *cur = list;
	while( cur )
	{
	    struct DTGStrList *next = cur->next;
	    if( !strcmp( cur->value, str ) )
	        return 0;
	    cur = next;
	}
	return 1;
}

DTGStrList *unmatch_DTGStrList( struct DTGStrList *list,
	            struct DTGStrList *mlist )
{
	struct DTGStrList *out = NULL;
	struct DTGStrList *cur = list;
	while( cur )
	{
	    struct DTGStrList *next = cur->next;
	    if( in_DTGStrList( mlist, cur->value ) != 0 )
	        out = append_DTGStrList( out, cur->value );
	    cur = next;
	}
	return out;
}

void print_DTGField( const struct DTGField *list )
{
	struct DTGField *cur = (DTGField*)list;
	while( cur )
	{
	    struct DTGField *next = cur->next;
	    if( cur->value )
	        printf( "cur->name:  \"%s\"\ncur->value:  \"%s\"\n\n",
	                cur->name, cur->value );
	    else
	        printf( "cur->name:  \"%s\"\ncur->value:  \"NULL\"\n\n",
	                cur->name );
	    cur = next;
	}
}

int count_DTGField( struct DTGField *list )
{
	struct DTGField *cur = list;
	int r = 0;
	while( cur )
	{
	    struct DTGField *next = cur->next;
	    r += 1;
	    cur = next;
	}
	return r;
}

void print_DTGFieldDesc( struct DTGFieldDesc *list )
{
	struct DTGFieldDesc *cur = list;
	while( cur )
	{
	    struct DTGFieldDesc *next = cur->next;
	    if( cur->type )
	        printf( "cur->name:  \"%s\"\ncur->type:  \"%s\"\n\n",
	                cur->name, cur->type );
	    if( strcmp( "select", cur->type ) == 0 )
	    {
	        print_DTGStrList( cur->select_values );
	        printf( "\n" );
	    }
	    cur = next;
	}
}

const char *sub_string( char *str, int start )
{
	str = str + start;
	return str;
}

struct DTGField *get_field( struct DTGField *fields, const char *id )
{
	if( !id )
	    return NULL;
	while( fields && strcmp( fields->name, id ) )
	    fields = fields->next;
	return fields;
}

const char *MyDTS::find_value( struct DTGField *fields, const char *id )
{
	if( !id )
	    return NULL;
	while( fields && strcmp( fields->value, id ) )
	    fields = fields->next;
	return fields ? fields->name : id;
}

const char *nullStr = "(null)";
const char *find_field( struct DTGField *fields, const char *id )
{
	fields = get_field( fields, id );
	if ( fields )
	{
	    if ( fields->value )
	        return fields->value;
	    else
	        return nullStr;
	}
	else
	    return id;
}

// SQL-escapes a string.  Does not enclose the string in surrounding quotes.
char* MyDTS::esc_field( const char* fld )
{
	if( !fld )
	    return NULL;

	unsigned long fldLen = strlen( fld );
	if( !fldLen )
	    return NULL;

	unsigned long bufSize = sizeof( char ) * fldLen * 2 + 1;
	char* buf = (char*) malloc( bufSize );

	mysql_real_escape_string( mysql, buf, fld, fldLen );

	return buf;
}

struct DTGField *MyDTS::single_row( const char *query, char *&err )
{
	struct DTGField *fields = NULL;
	if( !mysql_query( mysql, query ) )
	{
	    MYSQL_RES *res = mysql_store_result( mysql );
	    if( !res )
	        return fields;;
	    unsigned int f = mysql_num_fields( res );
	    MYSQL_FIELD *cols = mysql_fetch_fields( res );
	    MYSQL_ROW row = mysql_fetch_row( res );
	    for( unsigned int i = 0; row && i < f; i++ )
	        fields = append_DTGField( fields, new_DTGField(
	                        find_field( field_map, cols[i].name ),
	                        row[i] ) );
	    mysql_free_result( res );
	}
	else
	    err = mk_string( "Failed to retrieve data: ",
	                     mysql_error( mysql ) );
	return fields;
}

DTGStrList *
MyDTS::single_col( const char *query, char *&err )
{
	DTGStrList *values = NULL;

	if( !mysql_query( mysql, query ) )
	{
	    MYSQL_RES *res = mysql_store_result( mysql );
	    if( !res )
	        return values;
	    MYSQL_ROW row;
	    while( ( row = mysql_fetch_row( res ) ) )
	        values = append_DTGStrList( values, row[0] );
	    mysql_free_result( res );
	}
	else
	    err = mk_string( "Failed to retrieve data: ",
	                     mysql_error( mysql ) );

	return values;
}

char *MyDTS::get_description( const char *bugid, char *&err )
{
	char *query = mk_string("SELECT subject, description ",
	            "FROM issues WHERE id = ", bugid );

	char *desc = cp_string( "" );

	if( !mysql_query( mysql, query ) )
	{
	    MYSQL_RES *res = mysql_store_result( mysql );
	    if( !res )
	        return NULL;
	    MYSQL_ROW row;
	    int cnt = 0;
	    char cnttxt[10];
	    while( ( row = mysql_fetch_row( res ) ) )
	    {
	        sprintf( cnttxt, "%d", cnt++ );
	        char *tmp = mk_string( row[0], "\n\n", row[1] );
	        delete[] desc;
	        desc = tmp;
	    }
	    mysql_free_result( res );
	}
	else
	    err = mk_string( "Failed to retrieve description: ",
	                      mysql_error( mysql ) );

	delete[] query;

	return desc;
}

char *MyDTS::get_notes( const char *bugid, char *&err )
{
	char *query = mk_string( 
	    "SELECT journals.notes, journals.created_on, "
	    "users.login FROM journals, users WHERE "
	    "users.id=journals.user_id AND journalized_type='Issue' AND "
	    "journalized_id=", bugid, " ORDER BY created_on DESC;");

	char *notes = cp_string( "" );

	if( !mysql_query( mysql, query ) )
	{
	    MYSQL_RES *res = mysql_store_result( mysql );
	    if( !res )
	        return NULL;
	    MYSQL_ROW row;
	    int cnt = mysql_num_rows( res );
	    char cnttxt[10];

	    while( ( row = mysql_fetch_row( res ) ) )
	    {
	        sprintf( cnttxt, "%d", cnt-- );
	        if ( row[0] && strlen(row[0]) > 0)
	        {
	            char *tmp = mk_string( "--- Comment ",
	                    cnttxt, ": From ", row[2],
	                    " at ", row[1], "\n", row[0]);
	            char *tmp2 = mk_string( notes, tmp, "\n\n" );
	            delete[] tmp;
	            delete[] notes;
	            notes = tmp2;
	        }
	    }
	    mysql_free_result( res );
	}
	else
	    err = mk_string( "Failed to retrieve description: ",
	                      mysql_error( mysql ) );

	delete[] query;

	return notes;
}

int MyDTS::no_privs( char *&err )
{
	// These column/table names will never exist in the real schema, so if
	// we do have INSERT or UPDATE rights, then we won't change anything.

	const int acserr = 1142;

	const char *i = "INSERT INTO lajsglkjaal ( jASAS3a3a ) VALUES ( abcd )";
	if( mysql_query( mysql, i ) && mysql_errno( mysql ) == acserr )
	{
	    err = mk_string( "Check privileges: No 'INSERT' rights." );
	    return 1;
	}

	const char *u = "UPDATE laAHAAagw SET asAYWagsa = 1 WHERE alkWAj6j = 0";
	if( mysql_query( mysql, u ) && mysql_errno( mysql ) == acserr )
	{
	    err = mk_string( "Check privileges: No 'UPDATE' rights." );
	    return 1;
	}

	const char *s = "SELECT * FROM jlkasjfdlajsfdl";
	if( mysql_query( mysql, s ) && mysql_errno( mysql ) == acserr )
	{
	    err = mk_string( "Check privileges: No 'SELECT' rights." );
	    return 1;
	}
	return 0;
}

MyDTS::MyDTS( const char *server,
	    const char *user,
	    const char *pass,
	    const struct DTGField *attrs,
	    const char * /* prog_name */,
	    const char * /* prog_ver */,
	    char *&err)
{
	if( server )
	    my_server = cp_string( server );
	else
	    my_server = NULL;
	if( user )
	    my_user = cp_string( user );
	else
	    my_user = NULL;
	if( pass )
	    my_pass = cp_string( pass );
	else
	    my_pass = NULL;
	
	projID = NULL;
	use_profile = NULL;
	field_map = NULL;
	profile_map = NULL;
	product_map = NULL;
	component_map = NULL;

	mysql = NULL;
	if( err )
	{
	    valid = 0;
	    return;
	}

	// default MySQL port.
	my_port = 3306;

	// look for "host:port" format, otherwise use the default.  make sure
	// "server" only has one colon, and it's not at the beginning or end.
	const char* i0 =  strchr( my_server, ':' );
	const char* i1 = strrchr( my_server, ':' );

	if( i0                                            &&
	    i0 == i1                                      && // only one match
	    strlen( my_server ) != ( i0 - my_server + 1 ) && // not at end
	    i0 != my_server                                  // not at start
	  )
	    {
	        my_port = atoi( i0 + 1 );

	        char* s = my_server;
	        int end = i0 - s; // points to ':'.
	        my_server = new char[ end + 1 ];
	        my_server[ end ] = '\0';
	        strncpy( my_server, s, end );
	        delete[] s;
	    }

	struct DTGField *f;

	// run a privilege check?
	f = get_field( (DTGField*)attrs, "ck_privs" );
	if( f )
	    ck_privs = 0;
	else
	    ck_privs = 1;

	// SQL database name for Redmine
	f = get_field( (DTGField*)attrs, "rm_db" );
	if( f )
	    rm_db = cp_string( f->value );
	else
	    rm_db = cp_string( "redmine" );

	// custom field in Redmine that we use to store fix information in.
	f = get_field( (DTGField*)attrs, "rm_cf" );
	if( f )
	    rm_cf = cp_string( f->value );
	else
	    rm_cf = cp_string( "P4Fixes" );

	// Enable UNICODE?
	f = get_field( (DTGField*)attrs, "unicode" );
	if( f )
	    utf8 = 0;
	else
	    utf8 = 1;

	// Enable Fix replication with a 3rd party repository plug-in?
	f = get_field( (DTGField*)attrs, "use_repository" );
	if( f )
	    repo_repl = 0;
	else
	    repo_repl = 1;

	// set the offset
	f = get_field( (DTGField*)attrs, "offset" );
	if( f )
	    offset = atoi( f->value );
	else
	    offset = 0;

	f = get_field( (DTGField*)attrs, "wait_time" );
	if( f )
	    wait_time = atoi( f->value );
	else
	    wait_time = 10;

	mysql = mysql_init( NULL );

	// Try connecting
	valid = 0;
	valid = connected( err );
	if( err )
	    return;

	if( ck_privs )
	    if( no_privs( err ) )
	        return;

	char *qrm_db = esc_field( rm_db );
	char *uq = mk_string( "USE `", qrm_db, "`" );
	if( mysql_query( mysql, uq ) )
	{
	    err = mk_string( "Failed to USE database: \"", qrm_db, "\" - ",
	                          mysql_error( mysql ) );
	    delete[] uq;
	    free( qrm_db );
	    return;
	}
	free( qrm_db );
	delete[] uq;

	if( valid )
	    profile_map = two_cols( "SELECT id, login FROM users" );

	f = get_field( (DTGField*)attrs, "redmine_user" );
	if( f && strlen( f->value ) )    /* use the supplied Redmine account */
	{
	    const char* pmname = find_value( profile_map, f->value );
	    if( !strcmp( pmname, f->value ) )
	        err = mk_string( "\nFailed to find the supplied Redmine user ",
	                        "account.  Looked for \"", f->value, "\".\n" );
	    else
	        use_profile = cp_string( find_value( profile_map, pmname ) );
	}
	else    /* no specific Redmine user name set, so use the sql name */
	{
	    const char* pmname = find_value( profile_map, my_user );
	    if( !pmname || !strcmp( pmname, my_user ) )
	        err = mk_string( "\nFailed to find the same Redmine user ",
	            "name as that of your MySQL account.  Looked for ",
	            "\"", my_user, "\".\n", "\nTry the \"Redmine ",
	            "username\" field under the \"Edit ",
	            "attributes...\" button.\n" );
	    else
	        use_profile = cp_string( find_value( profile_map, my_user ) );
	}
}

MyDTS::~MyDTS()
{
	if( my_server )
	    delete[] my_server;
	if( my_user )
	    delete[] my_user;
	if( my_pass )
	    delete[] my_pass;
	if( use_profile )
	    delete[] use_profile;
	if( field_map )
	    delete_DTGField( field_map );
	if( profile_map )
	    delete_DTGField( profile_map );
	if( product_map )
	    delete_DTGField( product_map );
	if( component_map )
	    delete_DTGField( component_map );
	if( rm_db )
	  delete[] rm_db;
	if( rm_cf )
	  delete[] rm_cf;
	if( projID )
	  delete[] projID;
  
	// Close connections
	if( mysql )
	    mysql_close( mysql );
}

int
MyDTS::valid_project( const char *proj )
{
	if( !proj )
	    return 1;
	
	char *query = 
	    mk_string( "SELECT identifier FROM projects WHERE status=1" );
	if( !mysql_query( mysql, query ) )
	{
	    MYSQL_RES *res = mysql_store_result( mysql );
	    MYSQL_ROW row;
	    while( ( row = mysql_fetch_row( res ) ) )
	    {
	        if( !strcmp( row[0], proj ) )
	        {
	            mysql_free_result( res );
	            delete[] query;
	            return 0;
	        }
	    }
	    mysql_free_result( res );
	}
	delete[] query;
	return 1;
}

void
MyDTS::connect_to_project( const char *proj, char *&err )
{
	if ( !proj )
	    return;
	if ( projID )
	{
	    delete[] projID;
	    projID = NULL;
	}
	
	char *query = 
 	   mk_string( "SELECT id FROM projects WHERE identifier='", proj, "';");
	err = NULL;
	struct DTGStrList *tmp_list = single_col( query, err );
	if( !err && tmp_list )
	{
	    projID = cp_string( tmp_list->value );
	    delete_DTGStrList( tmp_list );
	}
	
	delete[] query;
}

int
MyDTS::connected( char *&err)
{
	// Check to see if connection is still valid
	if( valid )
	{
	    valid = !mysql_ping( mysql );
	    if( valid )
 	        return valid;
 	}
	valid = 0;

	// If the connection is invalid, try a new connection to the server.

	// Check for unrecoverable mysql_init() error
	if( !mysql )
	    return valid;

	int arg = MYSQL_PROTOCOL_TCP;
	mysql_options( mysql, MYSQL_OPT_PROTOCOL, (const char *)(&arg) );

	if( utf8 )
	    mysql_options( mysql, MYSQL_SET_CHARSET_NAME, "utf8" );

	MYSQL *tmp = mysql_real_connect( mysql, my_server, my_user, my_pass,
	    rm_db, my_port, NULL, 0 );
	if( !tmp )
	{
	    err = mk_string( "Connect Failure: ", mysql_error( mysql ) );
	    return valid;
	}

	return (valid = 1);
}

int
MyDTS::server_offline( struct DTGError *error )
{
	char *err = NULL;
	int ret = 0; // assume success

	if( !connected( err ) )
	{
	    set_DTGError( error, err );
	    ret = wait_time;
	}

	if( err )
	    delete[] err;

	return ret;
}

char *MyDTS::get_server_version( char *&err )
{
	if( !connected( err ) )
	    return NULL;
	char *ver = NULL;

	// Retrieve version from server

	const char *tmp = mysql_get_server_info( mysql );

	ver = mk_string( "MySQL:", tmp );

	return ver;
}

char *MyDTS::get_server_date( char *&err )
{
	if( !connected( err ) )
	    return NULL;

	char *date = NULL;

	// Retrieve date from server
	// Return a string which will be passed to extract_date to decode

	char query[128];
	sprintf( query, "%s%d%s", "SELECT ADDTIME(NOW(), -", offset, ");" );

	if( !mysql_query( mysql, query ) )
	{
	    MYSQL_RES *res = mysql_store_result( mysql );
	    MYSQL_ROW row = mysql_fetch_row( res );
	    if( row )
	        date = cp_string( row[0] );
	    else
	        err = mk_string( "Failed to retrieve row: ",
	                         mysql_error( mysql ) );
	    mysql_free_result( res );
	}
	else
	    err = mk_string( "Failed to retrieve date: ", mysql_error( mysql ));

	return date;
}

struct DTGStrList *MyDTS::list_projects( char *&err )
{
	if( !connected( err ) )
	    return NULL;

	struct DTGStrList *list = NULL;

	// Retrieve list of projects from server
	char *query = 
	    mk_string( "SELECT identifier FROM projects WHERE status=1" );
	if( !mysql_query( mysql, query ) )
	{
	    MYSQL_RES *res = mysql_store_result( mysql );
	    MYSQL_ROW row;
	    while( ( row = mysql_fetch_row( res ) ) )
	        list = append_DTGStrList( list, row[0] );
	    mysql_free_result( res );
	}
	else
	    err = mk_string( "Failed to get projects: ", mysql_error( mysql ) );
	delete[] query;
	if( !list )
	{
	    err = mk_string( "No projects available." );
	}
	return list;
}

struct DTGStrList *MyDTS::get_options( const char *table, char *&err )
{
	struct DTGStrList *list = NULL;
	char *query = mk_string( "SELECT * FROM ", table );
	if( !mysql_query( mysql, query ) )
	{
	    MYSQL_RES *res = mysql_store_result( mysql );
	    MYSQL_ROW row;
	    while( ( row = mysql_fetch_row( res ) ) )
	        list = append_DTGStrList( list, row[1] );
	    mysql_free_result( res );
	}
	else
	    err = mk_string( "Failed to get options: ", mysql_error( mysql ) );
	delete[] query;
	return list;
}

struct DTGStrList *MyDTS::get_custom_options( const char *options )
{
	struct DTGStrList *list = NULL;
	char delims[] = "\n";
	char *result = NULL;
	char *opts = cp_string(options);
	result = strtok( opts, delims );
	while( result != NULL ) 
	{
	    if (!strncmp(result, "- ", 2))
	        list = append_DTGStrList( list, strchr( result, result[2]));
	    result = strtok( NULL, delims );
	}
	delete[] opts;
	return list;
}

struct DTGStrList *MyDTS::get_prod_components( const char *product, char *&err )
{
	struct DTGStrList *list = NULL;
	char* qproduct = esc_field( product );
	char *query = mk_string( 
	    "SELECT issue_categories.name "
	    "FROM issue_categories, projects "
	    "WHERE issue_categories.project_id=projects.id "
	    "AND projects.name='", qproduct, "'" );

	delete[] qproduct;

	if( !mysql_query( mysql, query ) )
	{
	    MYSQL_RES *res = mysql_store_result( mysql );
	    MYSQL_ROW row;
	    while( ( row = mysql_fetch_row( res ) ) )
	        list = append_DTGStrList( list, row[0] );
	    mysql_free_result( res );
	}
	else
	    err = mk_string( "Failed to get categories: ",
	            mysql_error( mysql ) );

	delete[] query;
	return list;
}

struct DTGField *MyDTS::two_cols( const char *query )
{
	struct DTGField *list = NULL;
	if( !mysql_query( mysql, query ) )
	{
	    MYSQL_RES *res = mysql_store_result( mysql );
	    MYSQL_ROW row;
	    while( ( row = mysql_fetch_row( res ) ) )
	        list = append_DTGField( list, new_DTGField( row[0], row[1] ) );
	    mysql_free_result( res );
	}
	return list;
}

struct DTGField *MyDTS::field_names()
{
	struct DTGField *list = NULL;
	list = append_DTGField( list,
	        new_DTGField( "id", "IssueID" ) );
	list = append_DTGField( list,
	        new_DTGField( "tracker_id", "IssueType" ) );
	list = append_DTGField( list,
	        new_DTGField( "project_id", "Project" ) );
	list = append_DTGField( list,
	        new_DTGField( "subject", "Short description" ) );
	list = append_DTGField( list,
	        new_DTGField( "description", "Long Description" ) );
	list = append_DTGField( list,
	        new_DTGField( "due_date", "DueDate" ) );
	list = append_DTGField( list,
	        new_DTGField( "start_date", "StartDate" ) );
	list = append_DTGField( list,
	        new_DTGField( "category_id", "Category" ) );
	list = append_DTGField( list,
	        new_DTGField( "status_id", "Status" ) );
	list = append_DTGField( list,
	        new_DTGField( "assigned_to_id", "AssignedTo" ) );
	list = append_DTGField( list,
	        new_DTGField( "priority_id", "Priority" ) );
	list = append_DTGField( list,
	        new_DTGField( "fixed_version_id", "TargetVersion" ) );
	list = append_DTGField( list,
	        new_DTGField( "author_id", "ReportedBy" ) );
	list = append_DTGField( list,
	        new_DTGField( "updated_on", "UpdatedOn" ) );
	list = append_DTGField( list,
	        new_DTGField( "created_on", "CreatedOn" ) );
	list = append_DTGField( list,
	        new_DTGField( "lock_version", "NumberOfUpdates" ) );
	list = append_DTGField( list,
	        new_DTGField( "done_ratio", "DoneRatio" ) );
	list = append_DTGField( list,
	        new_DTGField( "estimated_hours", "EstimatedHours" ) );

	return list;
}

struct DTGFieldDesc *MyDTS::get_field_desc( const char *proj, char *&err )
{
	struct DTGFieldDesc *list = NULL;
	if( !field_map )
	    field_map = field_names();
	if( !mysql_query( mysql, "describe issues" ) )
	{
	    list = append_DTGFieldDesc( list,
	    new_DTGFieldDesc( "Description", "text", 1, NULL ) );
	    list = append_DTGFieldDesc( list,
	    new_DTGFieldDesc( "Notes", "text", 1, NULL ) );
	    list = append_DTGFieldDesc( list,
	    new_DTGFieldDesc( "Fixes", "fix", 0, NULL ) );
	        if (repo_repl == 0)
	        {
	            list = append_DTGFieldDesc( list,
	                new_DTGFieldDesc( "ChangeNums", "line", 0, NULL ) );
	        }
	    list = append_DTGFieldDesc( list,
	    new_DTGFieldDesc( "UpdatedBy", "word", 3, NULL ) );
	    MYSQL_RES *res = mysql_store_result( mysql );
	    MYSQL_ROW row;
	    while( ( row = mysql_fetch_row( res ) ) )
	    {
	        int rwflag = 1; // by default everything is ro
	        const char *type = NULL;
	        DTGStrList *options = NULL;
	        if( !strcmp( row[0], rm_cf ) )
	        {
	            rwflag = 0;
	            type = "line";
	        }
	        if( !strcmp( row[0], "id" ) )
	        {
	            rwflag = 4;
	            type = "word";
	        }
	        else if( !strcmp( row[0], "updated_on" ) )
	        {
	            rwflag = 2;
	            type = "date";
	        }
	        else if( !strcmp( row[0], "status_id" ) )
	        {
	            rwflag = 0;
	            type = "select";
	            options = get_options( "issue_statuses", err );
	        }
	        else if( !strcmp( row[0], "priority_id" ) )
	        {
	            rwflag = 0;
	            type = "select";
	            options = single_col( "SELECT name FROM enumerations "
	                    "WHERE type='IssuePriority';", err );
	        }
	        else if( !strcmp( row[0], "category_id" ) )
	        {
	            rwflag = 0;
	            type = "select";
	            char *query = mk_string("SELECT name FROM issue_categories "
	                    "WHERE project_id=", projID, ";");
	            options = single_col( query, err );
	            delete[] query;
	        }
	        else if( !strcmp( row[0], "project_id" ) )
	        {
	            rwflag = 1;
	            type = "line";
	        }
	        else if( !strcmp( row[0], "tracker_id" ) )
	        {
	            rwflag = 0;
	            type = "select";
	            char *query = mk_string("SELECT trackers.name FROM "
	                    "projects_trackers, trackers WHERE "
	                    "projects_trackers.tracker_id=trackers.id and "
	                    "projects_trackers.project_id='", projID, "'");
	            options = single_col( query, err );
	            delete[] query;
	        }
	        else
	        {
	            switch( row[1][0] )
	            {
	            case 'm': // mediumint, mediumtext
	                if( !strncmp( row[1], "mediumint(", 10 ) )
	                    type = "word";
	                else if( !strcmp( row[1], "mediumtext" ) )
	                {
	                    type = "text";
	                    rwflag = 0;
	                }
	                break;
	            case 't': // text, tinyint
	                if( !strcmp( row[1], "text" ) )
	                {
	                    type = "text";
	                    rwflag = 0;
	                }
	                else if( !strncmp( row[1], "tinyint(", 8 ) )
	                    type = "word";
	                break;
	            case 'd': // datetime, decimal
	                if( !strncmp( row[1], "datetime", 4 ) )
	                {
	                    type = "date";
	                    rwflag = 0;
	                }
	                else if( !strncmp( row[1], "decimal(", 8 ) )
	                    type = "word";
	                break;
	            case 'v': // varchar
	                if( !strncmp( row[1], "varchar(", 8 ) )
	                {
	                    type = "line";
	                    rwflag = 0;
	                }
	                break;
	            case 's': // smallint
	                if( !strncmp( row[1], "smallint(", 9 ) )
	                    type = "word";
	                break;
	            case 'i': // int
	                if( !strncmp( row[1], "int(", 4 ) )
	                    type = "word";
	                break;
	            case 'f': // float
	        if( !strcmp( row[1], "float" ) )
	                    type = "word";
	                break;
	            default:
	                break;
	            }
	        }

	        if( type )
	            list = append_DTGFieldDesc( list,
	        new_DTGFieldDesc( find_field( field_map, row[0] ),
	                    type, rwflag, options ) );
	    }
	    mysql_free_result( res );
	}
	else
	    err = mk_string( "Failed to get fields: ", mysql_error( mysql ) );

	    //Custom field time!
	    if( !mysql_query( mysql, 
	    "select name, field_format, possible_values, "
	        "editable from custom_fields where type='IssueCustomField'" ) )
	{
	        MYSQL_RES *res = mysql_store_result( mysql );
	    MYSQL_ROW row;
	    while( ( row = mysql_fetch_row( res ) ) )
	    {
	            int rwflag = 1; // by default everything is ro
	            const char *type = NULL;
	            DTGStrList *options = NULL;

	            switch( row[1][0] )
	            {
	            case 't': //text
	                rwflag = 0;
	                type = "text";
	                break;
	            case 'd': //date
	                type = "date";
	                break;
	            case 'l': //list
	                rwflag = 0;
	                type = "select";
	                options = get_custom_options( row[2] );
	                break;
	            case 's': //string
	                rwflag = 0;
	                type = "line";
	                break;
	            case 'i': //int
	            case 'f': //float
	            case 'b': //bool
	                type = "word";
	                break;
	            default:
	                break;
	            }

	            if ( row[3] == 0 )
	                rwflag = 1;

	            if( type )
	            {
	                char *tmp_string = mk_string( "custom_", row[0] );
	                list = append_DTGFieldDesc( list,
	                    new_DTGFieldDesc( tmp_string, 
						type, rwflag, options ) );
	                delete[] tmp_string;
	            }
	        }
	    mysql_free_result( res );
	}
	else
	    err = mk_string( "Failed to get fields: ", mysql_error( mysql ) );
	return list;
}

struct DTGStrList *
MyDTS::list_jobs( int max_rows,
	    struct DTGDate *since, const char *mod_date_field,
	    const char *exclude_user, const char *mod_by_field,
	    const char *seg_filters, char *&err )
{
	if( !connected( err ) )
	    return NULL;

	char q0[256];
	sprintf( q0, "%s %c%4.4d/%2.2d/%2.2d %2.2d:%2.2d:%2.2d%s%s%c",
	    "SELECT id FROM issues WHERE updated_on >=",
	    '"',
	    since->year, since->month, since->day,
	    since->hour, since->minute, since->second,
	    "\" AND project_id=\"", projID, '"'
	);
	char *query = mk_string( q0, seg_filters );
	struct DTGStrList *list = single_col( query, err );
	delete[] query;

	return list;
}

struct DTGField *MyDTS::get_defect( const char *defect, char *&err )
{
	// Initialize maps if needed
	if( !profile_map )
	    profile_map = two_cols( "SELECT id, login FROM users" );
	if( !field_map )
	    field_map = field_names();
	if( !product_map )
	    product_map = two_cols( "SELECT id, identifier FROM projects" );
	if( !component_map )
	    component_map = two_cols( "SELECT id, name FROM issue_categories" );

	// Retrieve specified defect and return the field values
	char* qdefect = esc_field( defect );
	char *query = mk_string( "SELECT * FROM issues WHERE id = ", qdefect );
	struct DTGField *result = single_row( query, err );

	struct DTGField *field = get_field( result, "AssignedTo" );
	if( field )
	{
	    if (field->value)
	    {
	        const char *tmp = find_field( profile_map, field->value );
	        free( field->value );
	        if( tmp )
	            field->value = strdup( tmp );
	        else
	            field->value = strdup( "NotFound" );
	    }
	    else
	        field->value = strdup( "NotSet" );
	}

	field = get_field( result, "ReportedBy" );
	if( field && field->value )
	{
	    const char *tmp = find_field( profile_map, field->value );
	    free( field->value );
	    if( tmp )
	        field->value = strdup( tmp );
	    else
	        field->value = strdup( "NotFound" );
	}

	field = get_field( result, "Project" );
	if( field && field->value )
	{
	    const char *tmp = find_field( product_map, field->value );
	    free( field->value );
	    if( tmp )
	        field->value = strdup( tmp );
	    else
	        field->value = strdup( "NotFound" );
	}

	field = get_field( result, "Category" );
	if( field )
	{
	    if (field->value)
	    {
	        const char *tmp = find_field( component_map, field->value );
	        free( field->value );
	        if( tmp )
	            field->value = strdup( tmp );
	        else
	            field->value = strdup( "" );
	    }
	    else
	    {
	        field->value = strdup( "" );
	    }
	}

	struct DTGField *tmp_field;
	field = get_field( result, "Status" );
	if( field && field->value )
	{
	    tmp_field = two_cols( "SELECT id, name FROM issue_statuses" );
	    const char *tmp = find_field( tmp_field, field->value );
	    free( field->value );
	    if( tmp )
	        field->value = strdup( tmp );
	    else
	        field->value = strdup( "NotFound" );
	    delete_DTGField( tmp_field );
	}

	field = get_field( result, "Priority" );
	if( field && field->value )
	{
	    tmp_field = two_cols( "SELECT id, name FROM enumerations" );
	    const char *tmp = find_field( tmp_field, field->value );
	    free( field->value );
	    if( tmp )
	        field->value = strdup( tmp );
	    else
	        field->value = strdup( "NotFound" );
	    delete_DTGField( tmp_field );
	}

	field = get_field( result, "IssueType" );
	if( field && field->value )
	{
	    tmp_field = two_cols( "SELECT id, name FROM trackers" );
	    const char *tmp = find_field( tmp_field, field->value );
	    free( field->value );
	    if( tmp )
	        field->value = strdup( tmp );
	    else
	        field->value = strdup( "NotFound" );
	    delete_DTGField( tmp_field );
	}

	field = get_field( result, "TargetVersion" );
	if( field && field->value )
	{
	    tmp_field = two_cols( "SELECT id, name FROM versions" );
	    const char *tmp = find_field( tmp_field, field->value );
	    delete_DTGField( tmp_field );
	    free( field->value );
	    if( tmp )
	        field->value = strdup( tmp );
	    else
	        field->value = strdup( "NotFound" );
	}
	delete[] query;
	if( !result )
	    err = cp_string( "Defect not found" );
	else
	{
	    char *desc = get_description( defect, err );
	    if( desc )
	        result = append_DTGField( result,
	            new_DTGField( "Description", desc ) );
	    delete[] desc;
	        char *notes = get_notes( defect, err );
	    if( desc )
	        result = append_DTGField( result,
	            new_DTGField( "Notes", notes ) );

	    delete[] notes;
	        //Get custom values
	        char *tmp_string = mk_string(
	                "SELECT custom_fields.name, custom_values.value ",
	                "FROM custom_values, custom_fields ",
	                "WHERE custom_values.customized_id=", qdefect,
	                " AND custom_values.customized_type = 'Issue' ",
	                "AND custom_values.custom_field_id=custom_fields.id" );
	        struct DTGField *tcustom = two_cols( tmp_string );
	        struct DTGField *custom = tcustom;
	        delete[] tmp_string;
	        while ( custom )
	        {
	            char *tmp_string = mk_string( "custom_", custom->name );
	            result = append_DTGField( result, new_DTGField(
						tmp_string, custom->value ) );
	            delete[] tmp_string;
	            custom = custom->next;
	        }
	        delete_DTGField( tcustom );
	}
	free( qdefect );
	return result;
}

void
MyDTS::replicate_fix( const char *defect, const char *nums, char *&err )
{
	// This function replicates fixes into Redmine.
	// If a 3rd party PErforce Repository plug-in is installed and in use

	if( !nums || !defect || repo_repl != 0 )
	    return;

	// If the ChangeNums field got in a weird state, it might not be set
	if ( !strcmp( nums, "ChangeNums" ) )
	    return;

	char* qdefect = esc_field( defect );
	char* query = NULL;

	// We need to parse and build a list of changelist numbers from the
	// ChangeNums field
	struct DTGStrList *changelist = NULL;
	char delims[] = " ";
	char *result = NULL;
	char *cchanges = cp_string(nums);
	result = strtok( cchanges, delims );
	while( result != NULL ) {
	    if( atoi( result ) != 0 )
	        changelist = append_DTGStrList( changelist, result );
	    result = strtok( NULL, delims );
	}
	delete[] cchanges;


	// Fetch the fixes that Redmine is already aware of
	// Redmine is SVN centric, so revision == changelist number
	query = mk_string(
	        "SELECT changesets.revision FROM changesets_issues, "
	        "changesets, repositories WHERE "
	        "changesets_issues.changeset_id=changesets.id AND "
	        "repositories.id=changesets.repository_id AND "
	        "changesets_issues.issue_id=", qdefect ,
	        " AND repositories.project_id=", projID );
	struct DTGStrList *fixes = single_col( query, err );
	delete[] query;

	if( err )
	    return;


	// Now we need to find the inersections of the two lists to find fixes
	// that need to be added and fiexes that need to be removed
	struct DTGStrList *tremove = unmatch_DTGStrList( fixes, changelist );
	struct DTGStrList *tadd = unmatch_DTGStrList( changelist, fixes );

	struct DTGStrList *remove = tremove;
	struct DTGStrList *add = tadd;

	delete_DTGStrList( changelist );
	delete_DTGStrList( fixes );

	while ( add )
	{
	    query = mk_string(
	    "INSERT INTO changesets_issues (changeset_id, "
	    "issue_id) VALUES ((SELECT changesets.id FROM changesets, "
	    "repositories WHERE changesets.revision=",
	    add->value, " AND repositories.id=changesets.repository_id AND "
	    "repositories.project_id=", projID, " ), ", qdefect, ")" );

	    // We don't care if this fails; that just means that there is
	    // missing data that the repository plug-in should have handled
	    mysql_query( mysql, query );
	    delete[] query;

	    add = add->next;
	}

	while ( remove )
	{
	    query = mk_string(
	    "DELETE FROM changesets_issues WHERE changeset_id=("
	    "SELECT changesets.id FROM changesets, repositories WHERE "
	    "changesets.revision=", add->value,
	    " AND repositories.id=changesets.repository_id AND "
	    "repositories.project_id=", projID, ") AND issue_id=",
	    qdefect );

	    // We don't care if this fails; that just means that there is
	    // missing data that the repository plug-in should have handled
	    mysql_query( mysql, query );
	    delete[] query;
	
	    remove = remove->next;
	}

	free( qdefect);
	delete_DTGStrList( tremove );
	delete_DTGStrList( tadd );
}

char *
MyDTS::save_defect( const char *defect, struct DTGField *fields, char *&err )
{
	if( !fields )
	    return cp_string( defect );

	struct DTGField *tfields = copy_DTGField( fields );
	char* qdefect = esc_field( defect );

	/* Build up SET clause list */
	struct DTGStrList *setlist = NULL;
	struct DTGStrList *sqlcmds = NULL;

	struct DTGStrList *timestamp = single_col( "SELECT NOW()", err );
	struct DTGStrList *journalid = NULL;

	// We need to create a journal if at least one field has changed.
	// Unless that field is ChangeNums
	const char *cNums = find_field( tfields, "ChangeNums" );
	if( count_DTGField(tfields) > 1 ||
	        ( count_DTGField( tfields ) == 1 &&
	          !strcmp( cNums, "ChangeNums" ) ) )
	{
	    const char *note = find_field( tfields, "Notes" );
	    const char *fix = find_field( tfields, "Fixes" );
	    if( fix[0] == '\n' )
	        fix = &fix[1];

	    // If there are notes or fix details to put into the journal,
	    // we need to do that now
	    if( strcmp( note, "Notes" ) || strcmp( fix, "Fixes" ) )
	    {
	        // Combine them, or use which-ever we have
	        char* notecombo = NULL;
	        if( strcmp( note, "Notes" ) && strcmp( fix, "Fixes" ) )
	            notecombo = mk_string( note, "\n\n", fix );
	        else if( strcmp( note, "Notes" ) )
	            notecombo = cp_string( note );
	        else
	            notecombo = cp_string( fix );

	        char *q0 = mk_string(
	                 "INSERT INTO journals (journalized_id, journalized_type, "
	                 "user_id, notes, created_on) VALUES (", qdefect,
	                ", \"Issue\", ", use_profile, ", \"" );
	        char *qnotecombo = esc_field( notecombo );
	        char *query = mk_string( q0, qnotecombo, "\", \"",
	            timestamp->value, "\")" );
	        free( qnotecombo );
	        if( !mysql_query( mysql, query) )
	        {

	            int cnt = mysql_affected_rows( mysql );
	            if( cnt < 0 )
	                err = mk_string( "SQL failed: make journal" );
	            else if( cnt > 1 )
	                err = mk_string("Too many rows updated: make journal");
	            else if( cnt == 0 )
	                err = mk_string( "Nothing happened: make journal" );
	        }
	        
	        delete[] q0;
	        delete[] query;
	        delete[] notecombo;
	    }
	    else
	    {
	        // We have no notes, so leave it blank
	        char *query = mk_string(
	                "INSERT INTO journals (journalized_id, journalized_type, "
	                "user_id, created_on) VALUES (", qdefect, ", \"Issue\", ",
	                use_profile, ", \"", timestamp->value, "\")" );
	        if( !mysql_query( mysql, query ) )
	        {
	            int cnt = mysql_affected_rows( mysql );
	            if( cnt < 0 )
	                err = mk_string( "SQL failed: make journal" );
	            else if( cnt > 1 )
	                err = mk_string("Too many rows updated: make journal");
	            else if( cnt == 0 )
	                err = mk_string( "Nothing happened: make journal" );
	        }
	        delete[] query;
	    }
	    
	    // If no errors encountered, we can fetch are new joural ID
	    if( !err )
	    {
	        char *query = mk_string(
	                "SELECT id FROM journals WHERE journalized_id=", qdefect,
	                " AND journalized_type=\"Issue\" AND created_on=\"",
	                timestamp->value, "\"" );
	        journalid = single_col( query, err );
	        delete[] query;
	    }

	    if( err )
	    {
	        free( qdefect );
	        delete_DTGStrList( timestamp );
	        delete_DTGStrList( journalid );
	        delete_DTGField( tfields );
	        return cp_string( defect );
	    }
	}


	// Now we have to disembiguate the other fields:
	// many of them need to be resolved via lookup tables
	for( struct DTGField *f = tfields; f; f = f->next )
	{
	    // First off, ignore it if it's special
	    const char *rn = find_value( field_map, f->name );
	    if( !strcmp( f->name, "Fixes" ) ||
	        !strcmp( f->name, "Notes" ) ||
	        !strcmp( f->name, "ChangeNums" ) ||
	        !strcmp( f->name, "description" ) ||
	        !strcmp( rn, "id" ) )
	        continue;

	    if ( !strcmp( f->name, "AssignedTo" ) )
	    {
	        const char *test = find_value(profile_map, f->value);
	        if ( strcmp(test, f->value) )
	        {
	            free(f->value);
	            f->value = strdup(test);
	        }
	        else
	        {
	            err = mk_string( "Failed to find option value: ",
	            f->name, " : ", f->value );
	        }
	    }

	    if ( !strcmp( f->name, "UpdatedBy" ) )
	    {
	        const char *test = find_value(profile_map, f->value);
	        if ( strcmp(test, f->value) )
	        {
	            free(f->value);
	            f->value = strdup(test);
	        }
	        else
	        {
	            err = mk_string( "Failed to find option value: ",
	                            f->name, " : ", f->value );
	        }
	    }

	    if ( !strcmp( f->name, "IssueType" ) )
	    {
	        struct DTGField *tmp_field = 
	            two_cols( "SELECT id, name FROM trackers;" );
	        const char *test = find_value( tmp_field, f->value);
	        if ( strcmp(test, f->value) )
	        {
	            free(f->value);
	            f->value = strdup(test);
	        }
	        else
	        {
	            err = mk_string( "Failed to find option value: ",
	                                f->name, " : ", f->value );
	        }
	        delete_DTGField( tmp_field );
	    }

	    if ( !strcmp( f->name, "Status" ) )
	    {
	        struct DTGField *tmp_field = 
	            two_cols( "SELECT id, name FROM issue_statuses;" );
	        const char *test = find_value( tmp_field, f->value);
	        if ( strcmp(test, f->value) )
	        {
	            free(f->value);
	            f->value = strdup(test);
	        }
	        else
	        {
	            err = mk_string( "Failed to find option value: ",
	                                f->name, " : ", f->value );
	        }
	        delete_DTGField( tmp_field );
	    }

	    if ( !strcmp( f->name, "Category" ) )
	    {
	        const char *test = find_value(component_map, f->value);
	        if ( strcmp(test, f->value) )
	        {
	            free(f->value);
	            f->value = strdup(test);
	        }
	        else
	        {
	            err = mk_string( "Failed to find option value: ",
	                            f->name, " : ", f->value );
	        }
	    }
	    
	    if ( !strcmp( f->name, "Priority" ) )
	    {
	        struct DTGField *tmp_field = 
	            two_cols( "SELECT id, name FROM enumerations;" );
	        const char *test = find_value( tmp_field, f->value);
	        if ( strcmp(test, f->value) )
	        {
	            free(f->value);
	            f->value = strdup(test);
	        }
	        else
	        {
	            err = mk_string( "Failed to find option value: ",
	                            f->name, " : ", f->value );
	        }
	        delete_DTGField( tmp_field );
	    }
	
	    if ( !strcmp( f->name, "TargetVersion" ) )
	    {
	        char *tmp_string = mk_string(
	                    "SELECT id, name FROM versions WHERE project_id=",
	                    projID, ";" );
	        struct DTGField *tmp_field = two_cols( tmp_string );
	        delete[] tmp_string;
	        const char *test = find_value( tmp_field, f->value);
	        if ( strcmp(test, f->value) )
	        {
	            free(f->value);
	            f->value = strdup(test);
	        }
	        else
	        {
	            err = mk_string( "Failed to find option value: ",
	                            f->name, " : ", f->value );
	        }
	        delete_DTGField( tmp_field );
	    }

	    if( err )
	    {
	        delete_DTGStrList( journalid );
	        delete_DTGStrList( setlist );
	        delete_DTGStrList( sqlcmds );
	        delete_DTGStrList( timestamp );
	        delete_DTGField( tfields );
	        free( qdefect );
	        return cp_string( defect );
	    }

	    if ( !strncmp( f->name, "custom_", 7 ) )
	    {
	        // This is a custom Redmine field.
	        // We need to make sure we don't update the wrong table
	        char *q0 = mk_string("SELECT id, name FROM custom_fields WHERE "
	                            "type='IssueCustomField'");
	        const char *customrn = find_value( two_cols( q0 ),
	                            sub_string( f->name, 7 ) );
	        char *q1 = mk_string("SELECT id, value FROM custom_values WHERE "
	                    "custom_field_id=", customrn,
	                    " AND customized_type='Issue' AND customized_id=",
	                    qdefect );
	        struct DTGField *test = single_row( q1 , err );
	        delete[] q0;
	        delete[] q1;

	        if( err )
	        {
	            delete_DTGField( test );
	            delete_DTGStrList( journalid );
	            delete_DTGStrList( setlist );
	            delete_DTGStrList( sqlcmds );
	            delete_DTGStrList( timestamp );
	            delete_DTGField( tfields );
	            delete[] customrn;
	            free( qdefect );
	            return cp_string( defect );
	        }


	        // Construct SQL statements to update the custom fields table
	        const char *customtest = find_field( test, "IssueID" );
	        const char *old;
	        int change = 0;
	        if( strcmp( customtest, "IssueID" ) )
	        {
	            // Have an existing value, so journal the change of value
	            old = find_field( test, "value" );
	            if( !strcmp( old, "value" ) ) {
	                delete[] old;
	                old = mk_string( "" );
	            }

	            // If there's no change, there's nothing to do
	            if( strcmp( old, f->value ) )
	            {
	                char *qvalue = esc_field( f->value );
	                char *tmp_string = mk_string(
	                            "UPDATE custom_values ",
	                            "SET value=\"", qvalue,
	                            "\" WHERE id=", customtest );
	                sqlcmds = append_DTGStrList( sqlcmds, tmp_string );
	                delete[] tmp_string;
	                free( qvalue );
	                change++;
	            }
	        }
	        else
	        {
	            // There is no current value, so journal the change from ""
	            char *qvalue = esc_field( f->value );
	            char *tmp_string = mk_string(
	                       "INSERT custom_values (customized_type, "
	                       "customized_id, custom_field_id, value) ",
	                       "VALUES ('Issue', '", qdefect, "', '", customrn,
	                       "', '", qvalue, "' )" );
	            sqlcmds = append_DTGStrList( sqlcmds, tmp_string );
	            delete[] tmp_string;
	            free( qvalue );
	            old = mk_string( "" );
	            change++;
	        }

	        // This SQL insert journals the change of value, if any
	        if( change > 0 )
	        {
	            char *tmp_string1 = mk_string(
	                "INSERT journal_details (journal_id, property, "
	                "prop_key, old_value, value) ",
	                "VALUES (", journalid->value, ", 'cf', '");
	            char *qold = esc_field( old );
	            char *qvalue = esc_field( f->value );
	            char *tmp_string2 = mk_string(
	                tmp_string1, customrn, "', '", qold,
	                "', '", qvalue, "' )" );
	            sqlcmds = append_DTGStrList( sqlcmds, tmp_string2 );
	            delete[] tmp_string1;
	            delete[] tmp_string2;
	            free( qvalue );
	            free( qold );
	        }
	    
	        delete[] old;
	    }
	    else
	    {
	        // We need to journal the changes to any non-custom fields here
	        // Lets find out what the field is right now.
	        char *tmp_string = mk_string(
	             "SELECT ", rn, " FROM issues WHERE id=\"", qdefect, "\"" );
	        struct DTGField *field = single_row( tmp_string, err );
	        char *old = cp_string( find_field( field, f->name ) );
	        delete_DTGField( field );
	        delete[] tmp_string;
	    
	        if( err )
	        {
	            delete_DTGStrList( journalid );
	            delete_DTGStrList( setlist );
	            delete_DTGStrList( sqlcmds );
	            delete_DTGStrList( timestamp );
	            delete_DTGField( tfields );
	            free( qdefect );
	            delete[] old;
	            return cp_string( defect );
	        }
	    
	        char *old_value = NULL;
	        if( !strcmp( old, f->name ) || !strcmp( old, nullStr ) )
	        {
	            old_value = cp_string( "NULL" );
	        }
	        else
	        {
	            char *qold = esc_field( old );
	            old_value = mk_string( "'", qold, "'" );
	            free( qold );
	        }
	        delete[] old;
	    
	        char *new_value = NULL;
	        char *tmp;
	        if( f->value )
	        {
	            char *qnew = esc_field( f->value );
	            tmp = mk_string( rn, "=\"", qnew, "\"" );
	            new_value = mk_string( "'", qnew, "'" );
	            free( qnew );
	        }
	        else
	        {
	            tmp = mk_string( rn, "=NULL" );
	            new_value = cp_string( "NULL" );
	        }
	        setlist = append_DTGStrList( setlist, tmp );
	        char *tmp_string1 = mk_string(
	                    "INSERT journal_details (journal_id, property, "
	                    "prop_key, old_value, value) ",
	                    "VALUES (", journalid->value, ", 'attr', '");
	        char *tmp_string2 = mk_string( tmp_string1, rn, "', ",
	                    old_value, ", ", new_value, " )" );
	        sqlcmds = append_DTGStrList( sqlcmds, tmp_string2 );
	        delete[] tmp_string1;
	        delete[] tmp_string2;
	        delete[] tmp;
	        delete[] old_value;
	        delete[] new_value;
	    }
	}
	delete_DTGStrList( journalid );
	delete_DTGField( tfields );

	// Lets build the main Issue update command 
	if( setlist || sqlcmds )
	{
	    char *tmp_string = mk_string( "updated_on=\"", timestamp->value,
	            "\"" );
	    setlist = append_DTGStrList( setlist, tmp_string );
	    delete[] tmp_string;
	    tmp_string = mk_string( "lock_version=lock_version+1" );
	    setlist = append_DTGStrList( setlist, tmp_string );
	    delete[] tmp_string;
	    char *sets = join_DTGStrList( setlist, ", " );
	    delete_DTGStrList( setlist );
	    tmp_string = mk_string( "UPDATE issues SET ", sets, " WHERE id = ",
	            qdefect );
	    sqlcmds = append_DTGStrList( sqlcmds, tmp_string );
	    delete[] tmp_string;

	    free( sets );
	}
	free( qdefect );
	delete_DTGStrList( timestamp );

	// Run UPDATE
	DTGStrList* isqlcmds = sqlcmds;
	while( isqlcmds )
	{
	    if( !mysql_query( mysql, isqlcmds->value ) )
	    {
	        int cnt = mysql_affected_rows( mysql );
	        if( cnt < 0 )
	            err = mk_string( "SQL failed:", isqlcmds->value );
	        else if( cnt > 1 )
	            err = mk_string( "Too many rows updated:", isqlcmds->value);
	    }
	    else
	        err = mk_string( "save_defect failed: \"", 
	            isqlcmds->value, "\", ",
	            mysql_error( mysql ) );
	
	    if( err )
	    {
	        delete_DTGStrList( sqlcmds );
	        return cp_string( defect );
	    }
	
	    isqlcmds = isqlcmds->next;
	}
	delete_DTGStrList( sqlcmds );
	    
	replicate_fix( defect, find_field( fields, "ChangeNums" ), err );
	
	return cp_string( defect );
}
